/**
 *
 * @param defaultValue
 *
 * @deprecated Use `@DTO.value`, `@DTO.factory` and `@DTO.dto` instead
 */
declare function Property<T extends () => unknown>(defaultValue: T): PropertyDecorator;
declare function Property<T>(defaultValue: T): PropertyDecorator;

type Class<T> = new (...args: unknown[]) => T;

interface Default {
    value(value: unknown): PropertyDecorator;
    factory(factory: (...args: unknown[]) => unknown): PropertyDecorator;
    dto(dtoType: Class<unknown>): PropertyDecorator;
    date(): PropertyDecorator;
    date(format: string): PropertyDecorator;
    date(stamp: number): PropertyDecorator;
    date(stamp?: string | number): PropertyDecorator;
}

declare function DTOWrapper<T extends object>(): Class<T>;
declare const DTO: typeof DTOWrapper & Default;

type IBuilder<T> = {
    [k in keyof T]-?: ((arg: T[k]) => IBuilder<T>) & (() => T[k]);
} & {
    /**
     * Build the object and return it
     */
    build(): T;
    /**
     * Derive a new builder from the current builder, copying it's
     * current values into a new builder. Allows default values to
     * be set in a copy, which will not be modified by child copies.
     */
    derive(): IBuilder<T>;
    /**
     * Assign an arbitrary value to a property. The value
     * does not need to match a valid property for the interface
     * and can be used to introduce "unexpected" values for testing,
     * or coming from other object maps.
     * @param property the name of the property to set a value for
     * @param value the value to set
     */
    assign<K>(property: string, value: K): IBuilder<T>;
    /**
     * Append a value to an array property. If no array
     * has been defined one will be created. If one exists,
     * it will be reused.
     * @param property
     * @param value
     */
    append<K>(property: string, value: K): IBuilder<T>;
    /**
     * Attach a value to a sub-property of a property.
     * If the property does not exist, a new empty object will be
     * assigned to it.
     *
     * The value is appended to the child object indexed by the sub-property.
     * @param property The name of the property to attach key:value pairs to
     * @param subProperty The name of the sub-property to attach the value to
     * @param value The value to attach to the child objects property
     */
    attach<K>(property: string, subProperty: string, value: K): ClsBuilder<T>;
};
type ClsBuilder<T> = {
    [k in keyof T]-?: BuilderMethod<T, T[k]>;
} & {
    /**
     * Build the object and return it
     */
    build(): T;
    /**
     * Assign an arbitrary value to a property. The value
     * does not need to match a valid property for the interface
     * and can be used to introduce "unexpected" values for testing,
     * or coming from other object maps.
     * @param property the name of the property to set a value for
     * @param value the value to set
     */
    assign<K>(property: string, value: K): ClsBuilder<T>;
    /**
     * Append a value to an array property. If no array
     * has been defined one will be created. If one exists,
     * it will be reused.
     * @param property
     * @param value
     */
    append<K>(property: string, value: K): ClsBuilder<T>;
    /**
     * Attach a value to a sub-property of a property.
     * If the property does not exist, a new empty object will be
     * assigned to it.
     *
     * The value is appended to the child object indexed by the sub-property.
     * @param property The name of the property to attach key:value pairs to
     * @param subProperty The name of the sub-property to attach the value to
     * @param value The value to attach to the child objects property
     */
    attach<K>(property: string, subProperty: string, value: K): ClsBuilder<T>;
};
type BuilderMethod<TTarget, TType> = ((arg: TType) => ClsBuilder<TTarget>) & {
    value: TType;
};
type DtoBuilder<T> = Class<ClsBuilder<T>> & {
    /**
     * Construct a builder from a raw object or class instance,
     * inheriting any properties that have already been defined.
     * @param value /
     */
    fromRaw(value: Partial<T>): ClsBuilder<T>;
    /**
     * Constructs a DTO, prefilled with any default values
     * assigned to the DTO via either the `Property` decorator
     * or the `DTO.*` decorators.
     *
     * With property:
     *
     * ```ts
     * class FooDto {
     *  ＠Property(1)
     *  foo: number;
     *  ＠Property(()=> "factory string")
     *  bar: string;
     *  ＠Property(SomeDto)
     *  baz: SomeDto;
     * }
     * ```
     *
     * With DTO:
     *
     * ```ts
     * class FooDto {
     *  ＠DTO.value(1)
     *  foo: number;
     *  ＠DTO.factory(()=> "factory string")
     *  bar: string;
     *  ＠DTO.dto(SomeDto)
     *  baz: SomeDto;
     * }
     * ```
     */
    default(): T;
};
declare function Builder<T>(cls: Class<T>): DtoBuilder<T>;
declare function Builder<T>(): IBuilder<T>;

export { Builder, ClsBuilder, DTO, DtoBuilder, Property };
